!---------------------------------------------------------------------!
!                NIST AMR Benchmark Reference Solution                !
!                                                                     !
! This module contains procedures for manipulating reference          !
! solutions as part of the NIST Adaptive Mesh Refinement Benchmark    !
! Problems project.  Reference solutions are provided for some        !
! problems for which the exact solution is not known.  For many uses  !
! the reference solution can be used in lieu of the exact solution    !
! up to the accuracy of the reference solution.  The primary          !
! procedures are read_refsoln which reads a reference solution from   !
! a file, and evaluate_refsoln which returns the value of the         !
! reference solution and optionally first derivatives at a given set  !
! of points.  For further  information, see the web page given below. !
!                                                                     !
! This is public domain software.  It was produced as part of work    !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/amr-benchmark                              !
!                                                                     !
!---------------------------------------------------------------------!

module AMRrefsoln ! Version 1.2.0

!===============================================================================
!   Description of public entities

use containing_element
implicit none
private
public &

   refsoln_type, &

! Structure for a reference solution.  The internals are private.  Declare your
! variable, for example my_refsoln, with
! type(refsoln_type) :: my_refsoln

   create_refsoln, &

! subroutine create_refsoln(refsoln)
! type(refsoln_type), intent(out) :: refsoln
! Initialize a refsoln variable. This must be called before any other operations
! are applied to refsoln.

   destroy_refsoln, &

! subroutine destroy_refsoln(refsoln,errcode)
! type(refsoln_type), intent(inout) :: refsoln
! integer, intent(out) :: errcode
! Destroy a refsoln variable.  This should be called to free all memory
! associated with refsoln.  No other operations may be applied to refsoln after
! calling destroy_refsoln, unless create_refsoln is called again first.
! errcode: 0 success
!          1 refsoln has not been created
!          2xx error code xx returned by ce_destroy

   read_refsoln, &

! subroutine read_refsoln(refsoln,filename,errcode)
! type(refsoln_type), intent(inout) :: refsoln
! character(len=*), intent(in) :: filename
! integer, intent(out) :: errcode
! Read the reference solution in the file given by filename.  This can only
! be called once between calls to refsoln_create and refsoln_destroy.
! errcode: 0 success
!          1 could not open filename
!          2 filename does not appear to be a NIST AMR Benchmark Reference
!            Solution file
!          3 unrecognized version number in filename
!          4 bad value in filename
!          5 other problem reading filename
!          6 allocation error
!          7 refsoln has not been created
!          8 refsoln has already been read
!          1xx error code xx returned by ce_create
!          3xx error code xx returned by ce_add_vertices
!          4xx error code xx returned by ce_add_elements
!          5xx error code xx returned by ce_make_search_data

   evaluate_refsoln, &

! subroutine evaluate_refsoln(refsoln,x,u,errcode,ux,uy,uz,soln_id)
! type(refsoln_type), intent(in) :: refsoln
! real(refsoln_real), intent(in) :: x(:) OR x(:,:)
! real(refsoln_real), intent(out) :: u OR u(:)
! integer, intent(out) :: errcode
! real(refsoln_real), intent(out), optional :: ux or ux(:)
! real(refsoln_real), intent(out), optional :: uy or uy(:)
! real(refsoln_real), intent(out), optional :: uz or uz(:)
! integer, intent(in), optional :: soln_id
! Evaluate the reference solution, and optionally first derivatives, at the
! given point(s).  The extent of the first dimension of x must be 1, 2 or 3 and
! must agree with the dimension of the solution in refsoln.  x(:) gives the
! coordinates of a single point, (x,y,z) with x(1)=x, x(2)=y and x(3)=z as
! appropriate for the dimension of the solution.  If x has rank 2 it contains
! a collection of points with x(:,i) giving the coordinates of the i'th point.
! The solution(s) is returned in u.  If a point is not contained in the domain
! of u, 0.0 is returned.  If x has rank 1, u must be a scalar.  If x has rank
! 2, u must have rank 1 and the extent of its dimension must be the same as the
! extent of the second dimension of x.  Optionally the first derivates are
! returned in ux, uy and uz. Any combination of them may be present, as
! appropriate for the dimension of the solution.  They follow the same rules as
! u, described above.
! soln_id is the ID of the solution for reference solutions that have more
! than one solution, for example which solution in a coupled system of
! equations.  The default, if not present, is 1.
! errcode: 0 success
!          1 extent of the first dimension of x does not agree with the
!            dimension in refsoln
!          2 rank or extent of u, ux, uy or uz does not agree with x
!          3 soln_id is out of range
!          4 requested derivative for a dimension that does not exist
!          5 degenerate element
!          6 refsoln has not been created
!          7 reference solution file has not been read
!          6xx error code -xx returned by ce_find_containing_element

   set_verbosity_refsoln, &

! subroutine set_verbosity_refsoln(refsoln,verbosity,errcode)
! type(refsoln_type), intent(inout) :: refsoln
! integer, intent(in) :: verbosity
! integer, intent(out) :: errcode
! Determines how much printed output to standard output occurs during the
! execution of AMRrefsoln procedures with argument refsoln.
! verbosity: 0 print nothing, not even errors and warnings
!            1 print errors, but not warning
!            2 print errors and warnings
!            3 print errors, warning and summary information about refsoln
!            4 print trace information for debugging
!            5 print detailed information for debugging
! The default is 2.
! errcode: 0 success
!          1 verbosity value out of range
!          2 refsoln has not been created

   get_refsoln_energy_errest, &

! function get_refsoln_energy_errest(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! real(refsoln_real) :: get_refsoln_energy_errest
! Return the estimate of the energy norm of the error.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_Linf_errest, &

! function get_refsoln_Linf_errest(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! real(refsoln_real) :: get_refsoln_Linf_errest
! Return the estimate of the L infinity norm of the error.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_L2_errest, &

! function get_refsoln_L2_errest(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! real(refsoln_real) :: get_refsoln_L2_errest
! Return the estimate of the L^2 norm of the error.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_H1_errest, &

! function get_refsoln_H1_errest(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! real(refsoln_real) :: get_refsoln_H1_errest
! Return the estimate of the H^1 norm of the error.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_energy_solution, &

! function get_refsoln_energy_solution(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! real(refsoln_real) :: get_refsoln_energy_solution
! Return the energy norm of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_Linf_solution, &

! function get_refsoln_Linf_solution(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! real(refsoln_real) :: get_refsoln_Linf_solution
! Return the L infinity norm of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_L2_solution, &

! function get_refsoln_L2_solution(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! real(refsoln_real) :: get_refsoln_L2_solution
! Return the L^2 norm of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_H1_solution, &

! function get_refsoln_H1_solution(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! real(refsoln_real) :: get_refsoln_H1_solution
! Return the H^1 norm of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_version, &

! function get_refsoln_version(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! character(len=10) :: get_refsoln_version
! Return the reference solution format version number.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_dimension, &

! function get_refsoln_dimension(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! integer :: get_refsoln_dimension
! Return the dimension (1, 2 or 3) of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_degree, &

! function get_refsoln_degree(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! integer :: get_refsoln_degree
! Return the degree of the piecewise polynomials used for reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_number_of_solutions, &

! function get_refsoln_number_of_solutions(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! integer :: get_refsoln_number_of_solutions
! Return the number of solutions in reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_number_of_vertices, &

! function get_refsoln_number_of_vertices(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! integer :: get_refsoln_number_of_vertices
! Return the number of vertices in the mesh.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_number_of_nodes, &

! function get_refsoln_number_of_nodes(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! integer :: get_refsoln_number_of_nodes
! Return the number of nodes in the mesh, including vertices.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_number_of_elements, &

! function get_refsoln_number_of_elements(refsoln,errcode)
! type(refsoln_type), intent(in) :: refsoln
! integer, intent(out) :: errcode
! integer :: get_refsoln_number_of_elements
! Return the number of elements in the mesh.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

   get_refsoln_bounding_box, &

! subroutine get_refsoln_bounding_box(refsoln,xmin,xmax,errcode)
! type(refsoln_type), intent(in) :: refsoln
! real(refsoln_real), intent(out) :: xmin(:), xmax(:)
! integer, intent(out) :: errcode
! Return a bounding box for the reference solution domain.  The extent of xmin
! and xmax must be at least the dimension of the solution.  Any point (x,y)
! (with appropriate modification for 1D and 3D) in the reference solution
! domain satisfies xmin(1) <= x <= xmax(1) and xmin(2) <= y <= xmax(2).
! errcode: 0 success
!          1 size of xmin or xmax is less than dimension of solution
!          2 refsoln has not been created
!          3 reference solution file has not been read

   refsoln_real

! KIND parameter for any real variables passed to AMRrefsoln procedures.
! Variables, for example real_var, should be declared with
! real(refsoln_real) :: real_var

! Generic procedures

interface evaluate_refsoln
   module procedure evaluate_refsoln_scalar, evaluate_refsoln_array, &
                    evaluate_refsoln_bad1,   evaluate_refsoln_bad2
end interface

!===============================================================================
!   User defined constants and data types

! Double precision kind number

integer, parameter :: refsoln_real=kind(1.0d0)

!------------------
! The mesh data structure

type mesh_type

! Simple scalars

   integer :: nvert, & ! The number of vertices in the grid.
              nnode, & ! The number of nodes in the grid, including vertices.
              nelem    ! The number of elements in the grid.

! Vertex coordinates.  The extent of the first dimension is 1, 2 or 3 depending
! on the dimension of the solution, and designates x, y or z.  The second
! dimension is the vertex id.

   real(refsoln_real), allocatable :: vertex(:,:)

! Solutions at the nodes, i.e. coefficients for Lagrangian nodal basis.  The
! first subscript is the solution id, e.g. which solution in a coupled system
! of equations.  The second is the node id.

   real(refsoln_real), allocatable :: solution(:,:)

! The nodes of each elements.  Each list begins with the vertices.  The first
! subscript is the local (within the element) id of the node, the second is the
! element id.  The value is the node global id.

   integer, allocatable :: element_node(:,:)

! Bounding box; the extent is equal to the dimension of the solution with
! entries 1, 2 and 3 giving the range for x, y and z, as appropriate

   real(refsoln_real), allocatable :: xmin(:), xmax(:)

end type mesh_type

! End of mesh data structure
!------------------

!------------------
! The reference solution data structure

type refsoln_type
   private

! Simple scalars

   integer :: dimension, & ! The dimension of the solution, 1D, 2D or 3D
              degree,    & ! The degree of the piecewise polynomial solution
              nsoln        ! The number of solutions at each node

   character(len=10) :: version ! The reference solution file format version

   real(refsoln_real) :: energy_errest, & ! estimate of energy norm of error
                         Linf_errest,   & ! estimate of L infinity norm of error
                         L2_errest,     & ! estimate of L^2 norm of error
                         H1_errest        ! estimate of H^1 norm of error

   real(refsoln_real) :: energy_solution, & ! energy norm of reference solution
                         Linf_solution,   & ! L infinity norm of ref. solution
                         L2_solution,     & ! L^2 norm of reference solution
                         H1_solution        ! H^1 norm of reference solution

! The mesh over which the piecewise polynomial reference solution is defined.

   type(mesh_type) :: mesh

! The data structure for finding an element that contains a given point

   type(ce_type) :: containing_element_data

! Status variables

   logical :: created = .false., & ! Has object been created and not destroyed?
              read                 ! Has a data file been read into the object?
   integer :: verbosity = 2        ! How much output to stdout

end type refsoln_type

! End of reference solution data structure
!------------------

!===============================================================================
!   Public routines

contains

!          --------------
subroutine create_refsoln(refsoln)
!          --------------

type(refsoln_type), intent(out) :: refsoln

! Initialize a refsoln variable. This must be called before any other operations
! are applied to refsoln.

refsoln%created = .true.
refsoln%read = .false.

end subroutine create_refsoln

!          ---------------
subroutine destroy_refsoln(refsoln,errcode)
!          ---------------

type(refsoln_type), intent(inout) :: refsoln
integer, intent(out) :: errcode

! Destroy a refsoln variable.  This should be called to free all memory
! associated with refsoln.  No other operations may be applied to refsoln after
! calling destroy_refsoln, unless create_refsoln is called again first.

! errcode: 0 success
!          1 refsoln has not been created
!          2xx error code xx returned by ce_destroy

integer :: i, ce_err

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in destroy_refsoln: refsoln has not been created"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter destroy_refsoln"

if (allocated(refsoln%mesh%vertex)) deallocate(refsoln%mesh%vertex)
if (allocated(refsoln%mesh%solution)) deallocate(refsoln%mesh%solution)
if (allocated(refsoln%mesh%element_node)) deallocate(refsoln%mesh%element_node)

if (refsoln%read) then
   call ce_destroy(refsoln%containing_element_data,ce_err)
   if (ce_err /= 0) then
      errcode = ce_err + 200
      if (refsoln%verbosity > 0) then
         print *,"ERROR in destroy_refsoln: ce_destroy returned error code ", &
                 ce_err
      endif
      return
   endif
endif


refsoln%created   = .false.
refsoln%read      = .false.
if (refsoln%verbosity >= 4) print *,"leave destroy_refsoln"
refsoln%verbosity = 2

end subroutine destroy_refsoln

!          ------------
subroutine read_refsoln(refsoln,filename,errcode)
!          ------------

type(refsoln_type), intent(inout) :: refsoln
character(len=*), intent(in) :: filename
integer, intent(out) :: errcode

! Read the reference solution in the file given by filename.  This can only
! be called once between calls to refsoln_create and refsoln_destroy.

! errcode: 0 success
!          1 could not open filename
!          2 filename does not appear to be a NIST AMR Benchmark Reference
!            Solution file
!          3 unrecognized version number in filename
!          4 bad value in filename
!          5 other problem reading filename
!          6 allocation error
!          7 refsoln has not been created
!          8 refsoln has already been read
!          1xx error code xx returned by ce_create
!          3xx error code xx returned by ce_add_vertices
!          4xx error code xx returned by ce_add_elements
!          5xx error code xx returned by ce_make_search_data

integer :: iounit, astat, num_other_nodes, nodes_per_element, i, ce_err
character(len=40) :: first_line
character(len=80) :: comment_line

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 7
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: refsoln has not been created"
   endif
   return
endif
if (refsoln%read) then
   errcode = 8
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: refsoln has already been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter read_refsoln"

! Open filename

iounit = available_iounit()
if (iounit == -2) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in  read_refsoln: could not find an available i/o unit"
   endif
   return
endif
if (refsoln%verbosity >= 5) print *,"open unit ",iounit," as file ",trim(filename)
open(unit=iounit,file=trim(filename),action="READ",status="OLD",err=100)

! Check that this is a reference solution file

read(iounit,"(a40)",err=200) first_line
if (first_line /= "NIST AMR Benchmark Reference Solution") then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: file ",trim(filename)," does not appear to be a NIST AMR Benchmark Reference Solution file"
   endif
   return
endif

! Verify the version of the file is supported by this version of the module

read(iounit,*,err=200) refsoln%version
if (refsoln%version /= "1.0" .and. &
    refsoln%version /= "1.1" .and. &
    refsoln%version /= "1.2") then
   errcode = 3
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: reference solution version is not supported by this software"
   endif
   return
endif
if (refsoln%verbosity >= 5) print *,"version ",refsoln%version

! Read past comments.

read(iounit,"(A80)",err=200) comment_line
comment_line = adjustl(comment_line)       ! allows leading blanks
do while (comment_line(1:1) == "#")
   read(iounit,"(A80)",err=200) comment_line
   comment_line = adjustl(comment_line)
end do

! Error estimates.  Note that comment_line contains the input line with the
! error estimates, and that the H1 norm was added at version 1.1.

if (refsoln%version == "1.0") then
   read(comment_line,*,err=200) refsoln%energy_errest, refsoln%Linf_errest, &
                                refsoln%L2_errest
   if (refsoln%verbosity >= 5) print *,"error estimates", &
                               refsoln%energy_errest,refsoln%Linf_errest, &
                               refsoln%L2_errest
else
   read(comment_line,*,err=200) refsoln%energy_errest, refsoln%Linf_errest, &
                                refsoln%L2_errest, refsoln%H1_errest
   if (refsoln%verbosity >= 5) print *,"error estimates", &
                               refsoln%energy_errest,refsoln%Linf_errest, &
                               refsoln%L2_errest,refsoln%H1_errest
endif

! Norms of the reference solution.  Added at version 1.2.

select case(refsoln%version)
case("1.0", "1.1")
   refsoln%energy_solution = -1.0
   refsoln%Linf_solution = -1.0
   refsoln%L2_solution = -1.0
   refsoln%H1_solution = -1.0
case("1.2")
   read(iounit,*,err=200) refsoln%energy_solution, refsoln%Linf_solution,&
                          refsoln%L2_solution, refsoln%H1_solution
   if (refsoln%verbosity >= 5) print *,"norms of reference solution", &
                               refsoln%energy_solution,refsoln%Linf_solution, &
                               refsoln%L2_solution,refsoln%H1_solution
end select

! Solution type.

read(iounit,*,err=200) refsoln%dimension, refsoln%degree, refsoln%nsoln
if (refsoln%verbosity >= 5) print *,"dimension, degree, nsoln", &
                            refsoln%dimension, refsoln%degree, refsoln%nsoln

! Vertices

read(iounit,*,err=200) refsoln%mesh%nvert, num_other_nodes
refsoln%mesh%nnode = refsoln%mesh%nvert + num_other_nodes
if (refsoln%verbosity >= 5) print *,"nvert, num_other_nodes", &
                            refsoln%mesh%nvert, num_other_nodes

allocate(refsoln%mesh%vertex(refsoln%dimension,refsoln%mesh%nvert), &
         refsoln%mesh%solution(refsoln%nsoln,refsoln%mesh%nnode), &
         stat=astat)
if (astat /= 0) then
   errcode = 6
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: allocation error for vertex and solution"
   endif
   return
endif

do i=1,refsoln%mesh%nvert
   read(iounit,*,err=200) refsoln%mesh%vertex(:,i),refsoln%mesh%solution(:,i)
end do
do i=refsoln%mesh%nvert+1,refsoln%mesh%nnode
   read(iounit,*,err=200) refsoln%mesh%solution(:,i)
end do

! Elements

read(iounit,*,err=200) refsoln%mesh%nelem
if (refsoln%verbosity >= 5) print *,"nelem",refsoln%mesh%nelem

select case(refsoln%dimension)
case (1)
   nodes_per_element = refsoln%degree + 1
case (2)
   nodes_per_element = ((refsoln%degree+1)*(refsoln%degree+2))/2
case (3)
   nodes_per_element = &
      (refsoln%degree**3+6*refsoln%degree**2+11*refsoln%degree+6)/6
case default
   errcode = 4
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: dimension is not 1, 2 or 3"
   endif
   return
end select
if (refsoln%verbosity >= 5) print *,"nodes per element",nodes_per_element

allocate(refsoln%mesh%element_node(nodes_per_element,refsoln%mesh%nelem), &
         stat=astat)
if (astat /= 0) then
   errcode = 6
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: allocation error for element nodes"
   endif
   return
endif

do i=1,refsoln%mesh%nelem
   read(iounit,*,err=200) refsoln%mesh%element_node(:,i)
end do

! Set the bounding box

allocate(refsoln%mesh%xmin(refsoln%dimension), &
         refsoln%mesh%xmax(refsoln%dimension), stat=astat)
if (astat /= 0) then
   errcode = 6
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: allocation error for bounding box"
   endif
   return
endif

refsoln%mesh%xmin(1) = minval(refsoln%mesh%vertex(1,:))
refsoln%mesh%xmax(1) = maxval(refsoln%mesh%vertex(1,:))
if (refsoln%dimension > 1) then
   refsoln%mesh%xmin(2) = minval(refsoln%mesh%vertex(2,:))
   refsoln%mesh%xmax(2) = maxval(refsoln%mesh%vertex(2,:))
endif
if (refsoln%dimension > 2) then
   refsoln%mesh%xmin(3) = minval(refsoln%mesh%vertex(3,:))
   refsoln%mesh%xmax(3) = maxval(refsoln%mesh%vertex(3,:))
endif

! Create the containing element data structure

call ce_create(refsoln%containing_element_data,ce_err)
if (ce_err /= 0) then
   errcode = ce_err + 100
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: ce_create returned error code ",ce_err
   endif
   return
endif

! Add the vertices to the containing element data structure

select case (refsoln%dimension)
case (1)
! TEMP don't have 1D containing_element
!   call ce_add_vertices(refsoln%containing_element_data, &
!                        refsoln%mesh%nvert, &
!                        (/(i,i=1,refsoln%mesh%nvert)/), &
!                        ce_err, &
!                        refsoln%mesh%vertex(1,1:refsoln%mesh%nvert))
   print *,"ERROR: can't do 1D reference solution because don't have 1D containing element"
   stop
case (2)
   call ce_add_vertices(refsoln%containing_element_data, &
                        refsoln%mesh%nvert, &
                        (/(i,i=1,refsoln%mesh%nvert)/), &
                        ce_err, &
                        refsoln%mesh%vertex(1,1:refsoln%mesh%nvert), &
                        refsoln%mesh%vertex(2,1:refsoln%mesh%nvert))
case (3)
   call ce_add_vertices(refsoln%containing_element_data, &
                        refsoln%mesh%nvert, &
                        (/(i,i=1,refsoln%mesh%nvert)/), &
                        ce_err, &
                        refsoln%mesh%vertex(1,1:refsoln%mesh%nvert), &
                        refsoln%mesh%vertex(2,1:refsoln%mesh%nvert), &
                        refsoln%mesh%vertex(3,1:refsoln%mesh%nvert))
end select
if (ce_err /= 0) then
   errcode = ce_err + 300
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: ce_add_vertices returned error code ", &
              ce_err
   endif
   return
endif

! Add the elements to the containing element data structure

call ce_add_elements(refsoln%containing_element_data, &
                     refsoln%mesh%nelem, &
                     (/(i,i=1,refsoln%mesh%nelem)/), &
         refsoln%mesh%element_node(1:refsoln%dimension+1,1:refsoln%mesh%nelem),&
                     ce_err)
if (ce_err /= 0) then
   errcode = ce_err + 400
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: ce_add_elements returned error code ", &
              ce_err
   endif
   return
endif

! Create the search data for containing element

call ce_make_search_data(refsoln%containing_element_data,ce_err)
if (ce_err /= 0) then
   errcode = ce_err + 500
   if (refsoln%verbosity > 0) then
      print *,"ERROR in read_refsoln: ce_make_search_data returned error code ", &
              ce_err
   endif
   return
endif

! All done!

refsoln%read = .true.
close(iounit)
if (refsoln%verbosity >= 4) print *,"leave read_refsoln"
return

! Error conditions during i/o statements

100 continue
errcode = 1
if (refsoln%verbosity > 0) then
   print *,"ERROR in read_refsoln: could not open file ",filename
endif
return

200 continue
errcode = 5
if (refsoln%verbosity > 0) then
   print *,"ERROR in read_refsoln: an error occured while reading file ",filename
endif
return

end subroutine read_refsoln

!          -----------------------
subroutine evaluate_refsoln_scalar(refsoln,x,u,errcode,ux,uy,uz,soln_id)
!          -----------------------

type(refsoln_type), intent(in) :: refsoln
real(refsoln_real), intent(in) :: x(:)
real(refsoln_real), intent(out) :: u
integer, intent(out) :: errcode
real(refsoln_real), intent(out), optional :: ux
real(refsoln_real), intent(out), optional :: uy
real(refsoln_real), intent(out), optional :: uz
integer, intent(in), optional :: soln_id

! Evaluate the reference solution, and optionally first derivatives, at the
! given point(s).  The extent of the first dimension of x must be 1, 2 or 3 and
! must agree with the dimension of the solution in refsoln.  x(:) gives the
! coordinates of a single point, (x,y,z) with x(1)=x, x(2)=y and x(3)=z as
! appropriate for the dimension of the solution.  If x has rank 2 it contains
! a collection of points with x(:,i) giving the coordinates of the i'th point.
! The solution(s) is returned in u.  If a point is not contained in the domain
! of u, 0.0 is returned.  If x has rank 1, u must be a scalar.  If x has rank
! 2, u must have rank 1 and the extent of its dimension must be the same as the
! extent of the second dimension of x.  Optionally the first derivates are
! returned in ux, uy and uz. Any combination of them may be present, as
! appropriate for the dimension of the solution.  They follow the same rules as
! u, described above.
! soln_id is the ID of the solution for reference solutions that have more
! than one solution, for example which solution in a coupled system of
! equations.  The default, if not present, is 1.
! errcode: 0 success
!          1 extent of the first dimension of x does not agree with the
!            dimension in refsoln
!          2 rank or extent of u, ux, uy or uz does not agree with x
!          3 soln_id is out of range
!          4 requested derivative for a dimension that does not exist
!          5 degenerate element
!          6 refsoln has not been created
!          7 reference solution file has not been read
!          6xx error code -xx returned by ce_find_containing_element

integer :: elem

! In case of early return

u = 0
if (present(ux)) ux = 0
if (present(uy)) uy = 0
if (present(uz)) uz = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 6
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 7
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: refsoln has not been read"
   endif
   return
endif
if (size(x) /= refsoln%dimension) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: first dimension of x does not agree with dimension of the solution"
   endif
   return
endif
if (present(soln_id)) then
   if (soln_id < 1 .or. soln_id > refsoln%nsoln) then
      errcode = 3
      if (refsoln%verbosity > 0) then
         print *,"ERROR in evaluate_refsoln: soln_id is out of range"
      endif
      return
   endif
endif
if ((present(uy) .and. refsoln%dimension < 2) .or. &
    (present(uz) .and. refsoln%dimension < 3)) then
   errcode = 4
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: requested derivative for dimension that does not exist"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter evaluate_refsoln_scalar"

! Find the element that contains the point

select case (refsoln%dimension)
case(1)
! TEMP don't have 1D containing_element
!   elem = ce_find_containing_element(refsoln%containing_element_data,x(1))
   print *,"can't do 1D reference solution because don't have 1D containing element"
   stop
case(2)
   elem = ce_find_containing_element(refsoln%containing_element_data,x(1),x(2))
case(3)
   elem = ce_find_containing_element(refsoln%containing_element_data,x(1), &
                                     x(2),x(3))
end select
if (elem < -1) then
   errcode = 600 - elem
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: ce_find_containing_element returned error code ",elem
   endif
   return
endif

! If elem is -1, did not find a containing element, so return 0

if (elem == -1) then
   u = 0
   if (present(ux)) ux = 0
   if (present(uy)) uy = 0
   if (present(uz)) uz = 0
else

! Otherwise, call the evaluation routine for the dimension of this refsoln

   select case(refsoln%dimension)
   case (1)
      call evaluate_refsoln_1D(refsoln,x,elem,u,errcode,ux,soln_id)
   case (2)
      call evaluate_refsoln_2D(refsoln,x,elem,u,errcode,ux,uy,soln_id)
   case (3)
      call evaluate_refsoln_3D(refsoln,x,elem,u,errcode,ux,uy,uz,soln_id)
   end select
   if (errcode == 1) errcode = 5
endif

if (refsoln%verbosity >= 4) print *,"leave evaluate_refsoln_scalar"

end subroutine evaluate_refsoln_scalar

!          ----------------------
subroutine evaluate_refsoln_array(refsoln,x,u,errcode,ux,uy,uz,soln_id)
!          ----------------------

type(refsoln_type), intent(in) :: refsoln
real(refsoln_real), intent(in) :: x(:,:)
real(refsoln_real), intent(out) :: u(:)
integer, intent(out) :: errcode
real(refsoln_real), intent(out), optional :: ux(:)
real(refsoln_real), intent(out), optional :: uy(:)
real(refsoln_real), intent(out), optional :: uz(:)
integer, intent(in) , optional :: soln_id

! Evaluate the reference solution, and optionally first derivatives, at the
! given point(s).  The extent of the first dimension of x must be 1, 2 or 3 and
! must agree with the dimension of the solution in refsoln.  x(:) gives the
! coordinates of a single point, (x,y,z) with x(1)=x, x(2)=y and x(3)=z as
! appropriate for the dimension of the solution.  If x has rank 2 it contains
! a collection of points with x(:,i) giving the coordinates of the i'th point.
! The solution(s) is returned in u.  If a point is not contained in the domain
! of u, 0.0 is returned.  If x has rank 1, u must be a scalar.  If x has rank
! 2, u must have rank 1 and the extent of its dimension must be the same as the
! extent of the second dimension of x.  Optionally the first derivates are
! returned in ux, uy and uz. Any combination of them may be present, as
! appropriate for the dimension of the solution.  They follow the same rules as
! u, described above.
! soln_id is the ID of the solution for reference solutions that have more
! than one solution, for example which solution in a coupled system of
! equations.  The default, if not present, is 1.
! errcode: 0 success
!          1 extent of the first dimension of x does not agree with the
!            dimension in refsoln
!          2 rank or extent of u, ux, uy or uz does not agree with x
!          3 soln_id is out of range
!          4 requested derivative for a dimension that does not exist
!          5 degenerate element
!          6 refsoln has not been created
!          7 reference solution file has not been read
!          6xx error code -xx returned by ce_find_containing_element

integer :: point, elem

! In case of early return

u = 0
if (present(ux)) ux = 0
if (present(uy)) uy = 0
if (present(uz)) uz = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 6
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 7
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: refsoln has not been read"
   endif
   return
endif
if (size(x,dim=1) /= refsoln%dimension) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: first dimension of x does not agree with dimension of the solution"
   endif
   return
endif
if (size(x,dim=2) /= size(u)) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: size(u) is not equal to size(x,dim=2)"
   endif
   return
endif
if (present(ux)) then
   if (size(x,dim=2) /= size(ux)) then
      errcode = 2
      if (refsoln%verbosity > 0) then
         print *,"ERROR in evaluate_refsoln: size(ux) is not equal to size(x,dim=2)"
      endif
      return
   endif
endif
if (present(uy)) then
   if (size(x,dim=2) /= size(uy)) then
      errcode = 2
      if (refsoln%verbosity > 0) then
         print *,"ERROR in evaluate_refsoln: size(uy) is not equal to size(x,dim=2)"
      endif
      return
   endif
endif
if (present(uz)) then
   if (size(x,dim=2) /= size(uz)) then
      errcode = 2
      if (refsoln%verbosity > 0) then
         print *,"ERROR in evaluate_refsoln: size(uz) is not equal to size(x,dim=2)"
      endif
      return
   endif
endif
if (present(soln_id)) then
   if (soln_id < 1 .or. soln_id > refsoln%nsoln) then
      errcode = 3
      if (refsoln%verbosity > 0) then
         print *,"ERROR in evaluate_refsoln: soln_id is out of range"
      endif
      return
   endif
endif
if ((present(uy) .and. refsoln%dimension < 2) .or. &
    (present(uz) .and. refsoln%dimension < 3)) then
   errcode = 4
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: requested derivative for dimension that does not exist"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter evaluate_refsoln_array"

! For the i'th point, use the element that contains the (i-1)'th point for
! the quick check.  For the first point, use a negative number to skip checking.

elem = -2

! For each point ...

do point=1,size(x,dim=2)

! Find the element that contains the point

! Find the element that contains the point

   select case (refsoln%dimension)
   case(1)
! TEMP don't have 1D containing_element
!      elem = ce_find_containing_element(refsoln%containing_element_data, &
!                                        x(1,point))
      print *,"can't do 1D reference solution because don't have 1D containing element"
      stop
   case(2)
      elem = ce_find_containing_element(refsoln%containing_element_data, &
                                        x(1,point),x(2,point))
   case(3)
      elem = ce_find_containing_element(refsoln%containing_element_data, &
                                        x(1,point),x(2,point),x(3,point))
   end select
   if (elem < -1) then
      errcode = 600 - elem
      if (refsoln%verbosity > 0) then
         print *,"ERROR in evaluate_refsoln: ce_find_containing_element returned error code ",elem
      endif
      return
   endif

! If elem is -1, did not find a containing element, so return 0

   if (elem == -1) then
      u(point) = 0
      if (present(ux)) ux(point) = 0
      if (present(uy)) uy(point) = 0
      if (present(uz)) uz(point) = 0
   else

! Otherwise, call the evaluation routine for the dimension of this refsoln

      select case(refsoln%dimension)
      case (1)
         if (present(ux)) then
            call evaluate_refsoln_1D(refsoln,x(:,point),elem,u(point),errcode, &
                                     ux(point),soln_id)
         else
            call evaluate_refsoln_1D(refsoln,x(:,point),elem,u(point),errcode, &
                                     soln_id=soln_id)
         endif
      case (2)
         if (present(ux)) then
          if (present(uy)) then
            call evaluate_refsoln_2D(refsoln,x(:,point),elem,u(point),errcode, &
                                     ux(point),uy(point),soln_id)
          else
            call evaluate_refsoln_2D(refsoln,x(:,point),elem,u(point),errcode, &
                                     ux(point),soln_id=soln_id)
          endif
         else
          if (present(uy)) then
            call evaluate_refsoln_2D(refsoln,x(:,point),elem,u(point),errcode, &
                                     uy=uy(point),soln_id=soln_id)
          else
            call evaluate_refsoln_2D(refsoln,x(:,point),elem,u(point),errcode, &
                                     soln_id=soln_id)
          endif
         endif
      case (3)
         if (present(ux)) then
          if (present(uy)) then
           if (present(uz)) then
            call evaluate_refsoln_3D(refsoln,x(:,point),elem,u(point),errcode, &
                                     ux(point),uy(point),uz(point),soln_id)
           else
            call evaluate_refsoln_3D(refsoln,x(:,point),elem,u(point),errcode, &
                                     ux(point),uy(point),soln_id=soln_id)
           endif
          else
           if (present(uz)) then
            call evaluate_refsoln_3D(refsoln,x(:,point),elem,u(point),errcode, &
                                     ux(point),uz=uz(point),soln_id=soln_id)
           else
            call evaluate_refsoln_3D(refsoln,x(:,point),elem,u(point),errcode, &
                                     ux(point),soln_id=soln_id)
           endif
          endif
         else
          if (present(uy)) then
           if (present(uz)) then
            call evaluate_refsoln_3D(refsoln,x(:,point),elem,u(point),errcode, &
                                     uy=uy(point),uz=uz(point),soln_id=soln_id)
           else
            call evaluate_refsoln_3D(refsoln,x(:,point),elem,u(point),errcode, &
                                     uy=uy(point),soln_id=soln_id)
           endif
          else
           if (present(uz)) then
            call evaluate_refsoln_3D(refsoln,x(:,point),elem,u(point),errcode, &
                                     uz=uz(point),soln_id=soln_id)
           else
            call evaluate_refsoln_3D(refsoln,x(:,point),elem,u(point),errcode, &
                                     soln_id=soln_id)
           endif
          endif
         endif
      end select
      if (errcode == 1) errcode = 5
   endif

end do

if (refsoln%verbosity >= 4) print *,"leave evaluate_refsoln_array"

end subroutine evaluate_refsoln_array

!          ---------------------
subroutine evaluate_refsoln_bad1(refsoln,x,u,errcode,ux,uy,uz,soln_id)
!          ---------------------

type(refsoln_type), intent(in) :: refsoln
real(refsoln_real), intent(in) :: x(:)
real(refsoln_real), intent(out) :: u(:)
integer, intent(out) :: errcode
real(refsoln_real), intent(out), optional :: ux(:)
real(refsoln_real), intent(out), optional :: uy(:)
real(refsoln_real), intent(out), optional :: uz(:)
integer, intent(in), optional :: soln_id

! This is to catch the user error of passing rank 1 u with rank 1 x, rather
! than have the compiler give a possibly confusing "no matching specific
! routine" compile time error.

! In case of early return

u = 0
if (present(ux)) ux = 0
if (present(uy)) uy = 0
if (present(uz)) uz = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 3
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 4
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: refsoln has not been read"
   endif
   return
endif

! The rank of u does not agree with x

errcode = 2
if (refsoln%verbosity > 0) then
   print *,"ERROR in evaluate_refsoln: x has rank 1 so u must be a scalar"
endif

end subroutine evaluate_refsoln_bad1

!          ---------------------
subroutine evaluate_refsoln_bad2(refsoln,x,u,errcode,ux,uy,uz,soln_id)
!          ---------------------

type(refsoln_type), intent(in) :: refsoln
real(refsoln_real), intent(in) :: x(:,:)
real(refsoln_real), intent(out) :: u
integer, intent(out) :: errcode
real(refsoln_real), intent(out), optional :: ux(:)
real(refsoln_real), intent(out), optional :: uy(:)
real(refsoln_real), intent(out), optional :: uz(:)
integer, intent(in), optional :: soln_id

! This is to catch the user error of passing scalar u with rank 2 x, rather
! than have the compiler give a possibly confusing "no matching specific
! routine" compile time error.

! In case of early return

u = 0
if (present(ux)) ux = 0
if (present(uy)) uy = 0
if (present(uz)) uz = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 3
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 4
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: refsoln has not been read"
   endif
   return
endif

! The rank of u does not agree with x

errcode = 2
if (refsoln%verbosity > 0) then
   print *,"ERROR in evaluate_refsoln: x has rank 2 so u must have rank 1"
endif

end subroutine evaluate_refsoln_bad2

!          ---------------------
subroutine set_verbosity_refsoln(refsoln,verbosity,errcode)
!          ---------------------

type(refsoln_type), intent(inout) :: refsoln
integer, intent(in) :: verbosity
integer, intent(out) :: errcode

! Determines how much printed output to standard output occurs during the
! execution of AMRrefsoln procedures with argument refsoln.
! verbosity: 0 print nothing, not even errors and warnings
!            1 print errors, but not warning
!            2 print errors and warnings
!            3 print errors, warning and summary information about refsoln
!            4 print trace information for debugging
!            5 print detailed information for debugging
! The default is 2.
! errcode: 0 success
!          1 verbosity value out of range
!          2 refsoln has not been created

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in set_verbosity_refsoln: refsoln has not been created"
   endif
   return
endif
if (verbosity < 0 .or. verbosity > 5) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in set_verbosity_refsoln: verbosity out of range, value is ", &
              verbosity
   endif
   return
endif

if (refsoln%verbosity >= 4 .and. verbosity < 4) then
   print *,"reducing verbosity to stop trace"
endif
if (refsoln%verbosity < 4 .and. verbosity >= 4) then
   print *,"increasing verbosity to start trace"
endif

refsoln%verbosity = verbosity

end subroutine set_verbosity_refsoln

!        -------------------------
function get_refsoln_energy_errest(refsoln,errcode)
!        -------------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
real(refsoln_real) :: get_refsoln_energy_errest

! Return the estimate of the energy norm of the error.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_energy_errest = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_energy_errest: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_energy_errest: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_energy_errest"

get_refsoln_energy_errest = refsoln%energy_errest

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_energy_errest"

end function get_refsoln_energy_errest

!        -----------------------
function get_refsoln_Linf_errest(refsoln,errcode)
!        -----------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
real(refsoln_real) :: get_refsoln_Linf_errest

! Return the estimate of the L infinity norm of the error.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_Linf_errest = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_Linf_errest: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_Linf_errest: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_Linf_errest"

get_refsoln_Linf_errest = refsoln%Linf_errest

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_Linf_errest"

end function get_refsoln_Linf_errest

!        ---------------------
function get_refsoln_L2_errest(refsoln,errcode)
!        ---------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
real(refsoln_real) :: get_refsoln_L2_errest

! Return the estimate of the L^2 norm of the error.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_L2_errest = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_L2_errest: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_L2_errest: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_L2_errest"

get_refsoln_L2_errest = refsoln%L2_errest

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_L2_errest"

end function get_refsoln_L2_errest

!        ---------------------
function get_refsoln_H1_errest(refsoln,errcode)
!        ---------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
real(refsoln_real) :: get_refsoln_H1_errest

! Return the estimate of the H^1 norm of the error.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_H1_errest = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_H1_errest: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_H1_errest: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_H1_errest"

get_refsoln_H1_errest = refsoln%H1_errest

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_H1_errest"

end function get_refsoln_H1_errest

!        -------------------------
function get_refsoln_energy_solution(refsoln,errcode)
!        -------------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
real(refsoln_real) :: get_refsoln_energy_solution

! Return the energy norm of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_energy_solution = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_energy_solution: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_energy_solution: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_energy_solution"

get_refsoln_energy_solution = refsoln%energy_solution

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_energy_solution"

end function get_refsoln_energy_solution

!        -----------------------
function get_refsoln_Linf_solution(refsoln,errcode)
!        -----------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
real(refsoln_real) :: get_refsoln_Linf_solution

! Return the L infinity norm of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_Linf_solution = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_Linf_solution: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_Linf_solution: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_Linf_solution"

get_refsoln_Linf_solution = refsoln%Linf_solution

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_Linf_solution"

end function get_refsoln_Linf_solution

!        ---------------------
function get_refsoln_L2_solution(refsoln,errcode)
!        ---------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
real(refsoln_real) :: get_refsoln_L2_solution

! Return the L^2 norm of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_L2_solution = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_L2_solution: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_L2_solution: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_L2_solution"

get_refsoln_L2_solution = refsoln%L2_solution

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_L2_solution"

end function get_refsoln_L2_solution

!        ---------------------
function get_refsoln_H1_solution(refsoln,errcode)
!        ---------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
real(refsoln_real) :: get_refsoln_H1_solution

! Return the H^1 norm of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_H1_solution = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_H1_solution: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_H1_solution: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_H1_solution"

get_refsoln_H1_solution = refsoln%H1_solution

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_H1_solution"

end function get_refsoln_H1_solution

!        -------------------
function get_refsoln_version(refsoln,errcode)
!        -------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
character(len=10) :: get_refsoln_version

! Return the reference solution format version number.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_version = "0.0"

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_version: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_version: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_version"

get_refsoln_version = refsoln%version

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_version"

end function get_refsoln_version

!        ---------------------
function get_refsoln_dimension(refsoln,errcode)
!        ---------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
integer :: get_refsoln_dimension

! Return the dimension (1, 2 or 3) of the reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_dimension = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_dimension: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_dimension: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_dimension"

get_refsoln_dimension = refsoln%dimension

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_dimension"

end function get_refsoln_dimension

!        ------------------
function get_refsoln_degree(refsoln,errcode)
!        ------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
integer :: get_refsoln_degree

! Return the degree of the piecewise polynomials used for reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_degree = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_degree: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_degree: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_degree"

get_refsoln_degree = refsoln%degree

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_degree"

end function get_refsoln_degree

!        -------------------------------
function get_refsoln_number_of_solutions(refsoln,errcode)
!        -------------------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
integer :: get_refsoln_number_of_solutions

! Return the number of solutions in reference solution.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_number_of_solutions = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_number_of_solutions: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_number_of_solutions: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_number_of_solutions"

get_refsoln_number_of_solutions = refsoln%nsoln

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_number_of_solutions"

end function get_refsoln_number_of_solutions

!        ------------------------------
function get_refsoln_number_of_vertices(refsoln,errcode)
!        ------------------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
integer :: get_refsoln_number_of_vertices

! Return the number of vertices in the mesh.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_number_of_vertices = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_number_of_vertices: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_number_of_vertices: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_number_of_vertices"

get_refsoln_number_of_vertices = refsoln%mesh%nvert

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_number_of_vertices"

end function get_refsoln_number_of_vertices

!        ---------------------------------
function get_refsoln_number_of_nodes(refsoln,errcode)
!        ---------------------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
integer :: get_refsoln_number_of_nodes

! Return the number of nodes in the mesh, including vertices.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_number_of_nodes = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_number_of_nodes: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_number_of_nodes: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_number_of_nodes"

get_refsoln_number_of_nodes = refsoln%mesh%nnode

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_number_of_nodes"

end function get_refsoln_number_of_nodes

!        ------------------------------
function get_refsoln_number_of_elements(refsoln,errcode)
!        ------------------------------

type(refsoln_type), intent(in) :: refsoln
integer, intent(out) :: errcode
integer :: get_refsoln_number_of_elements

! Return the number of elements in the mesh.
! errcode: 0 success
!          1 refsoln has not been created
!          2 reference solution file has not been read

! In case of early return

get_refsoln_number_of_elements = 0

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_number_of_elements: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_number_of_elements: refsoln has not been read"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_number_of_elements"

get_refsoln_number_of_elements = refsoln%mesh%nelem

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_number_of_elements"

end function get_refsoln_number_of_elements

!          ------------------------
subroutine get_refsoln_bounding_box(refsoln,xmin,xmax,errcode)
!          ------------------------

type(refsoln_type), intent(in) :: refsoln
real(refsoln_real), intent(out) :: xmin(:), xmax(:)
integer, intent(out) :: errcode

! Return a bounding box for the reference solution domain.  The extent of xmin
! and xmax must be at least the dimension of the solution.  Any point (x,y)
! (with appropriate modification for 1D and 3D) in the reference solution
! domain satisfies xmin(1) <= x <= xmax(1) and xmin(2) <= y <= xmax(2).
! errcode: 0 success
!          1 size of xmin or xmax is less than dimension of solution
!          2 refsoln has not been created
!          3 reference solution file has not been read

! Check for input errors

errcode = 0
if (.not. refsoln%created) then
   errcode = 2
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_bounding_box: refsoln has not been created"
   endif
   return
endif
if (.not. refsoln%read) then
   errcode = 3
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_bounding_box: refsoln has not been read"
   endif
   return
endif
if (size(xmin) < refsoln%dimension .or. size(xmax) < refsoln%dimension) then
   errcode = 1
   if (refsoln%verbosity > 0) then
      print *,"ERROR in get_refsoln_bounding_box: size of xmin or xmax is less than dimension of solution"
   endif
   return
endif

if (refsoln%verbosity >= 4) print *,"enter get_refsoln_bounding_box"

xmin(1:refsoln%dimension) = refsoln%mesh%xmin
xmax(1:refsoln%dimension) = refsoln%mesh%xmax

if (refsoln%verbosity >= 4) print *,"leave get_refsoln_bounding_box"

end subroutine get_refsoln_bounding_box

!===============================================================================
!   Private routines

!        ----------------
function available_iounit()
!        ----------------

integer :: available_iounit

! This function returns an iounit that exists and is not in use.  If it
! cannot find one, it returns -2 as an error flag.

logical :: exists, opened

available_iounit = 11
do while (available_iounit < 1000)
   inquire(unit=available_iounit,exist=exists,opened=opened)
   if (exists .and. .not. opened) exit
   available_iounit = available_iounit + 1
end do
if (available_iounit == 1000) available_iounit = -2

end function available_iounit

!          -----------
subroutine barycentric(x,y,xvert,yvert,zeta,errcode,zetax,zetay,no_det)
!          -----------

!----------------------------------------------------
! This subroutine returns the barycentric coordinates of (x,y)
! in the triangle with vertices (xvert,yvert).
! If no_det is present and .true., the coordinates are not divided by the
! determinant (area of the triangle).  This is useful if you are only looking
! at the signs of the coordinates, where it can be wrong if the point is
! very close to an edge and the area is very small so roundoff can be very bad.
!
! errcode: 0 success
!          1 degenerate element
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

real(refsoln_real), intent(in) :: x,y,xvert(:),yvert(:)
real(refsoln_real), intent(out) :: zeta(:)
integer, intent(out) :: errcode
real(refsoln_real), intent(out), optional :: zetax(:), zetay(:)
logical, optional, intent(in) :: no_det

!----------------------------------------------------
! Local variables:

! If your compiler reports that quad_real is not a legal value for a kind
! type parameter, then it does not support quadruple precision (16 byte) reals.
! In that case, use quad_real = selected_real_kind(p=14) and hope the reference
! solution doesn't have any very tiny triangles.

integer, parameter :: quad_real = selected_real_kind(p=30)

real(quad_real) :: x1,x2,x3,y1,y2,y3,det,ssump,ssumm
real(quad_real) :: xy1,xy2,xy3,yx1,yx2,yx3,sump,summ
real(quad_real) :: x1y2,x2y3,x3y1,x1y3,x2y1,x3y2
logical :: local_no_det

!----------------------------------------------------
! Begin executable code

! check for request for no determinant

if (present(no_det)) then
   local_no_det = no_det
else
   local_no_det = .false.
endif

! put the vertices in local variables to shorten expressions

x1 = real(xvert(1),quad_real)
x2 = real(xvert(2),quad_real)
x3 = real(xvert(3),quad_real)
y1 = real(yvert(1),quad_real)
y2 = real(yvert(2),quad_real)
y3 = real(yvert(3),quad_real)

! compute the barycentric coordinates of the point

! reduce roundoff by summing all the positive parts and negative parts
! separately, and then adding the two partial sums

! products needed for the sums

xy1 = real(x,quad_real)*y1
xy2 = real(x,quad_real)*y2
xy3 = real(x,quad_real)*y3
yx1 = real(y,quad_real)*x1
yx2 = real(y,quad_real)*x2
yx3 = real(y,quad_real)*x3
x1y2 = x1*y2
x2y3 = x2*y3
x3y1 = x3*y1
x1y3 = x1*y3
x2y1 = x2*y1
x3y2 = x3*y2

! compute the determinant

! det = x1*(y2-y3)+x2*(y3-y1)+x3*(y1-y2)

ssump=0.0_quad_real; ssumm=0.0_quad_real
if (x1y2 > 0.0_quad_real) then; ssump=ssump+x1y2; else; ssumm=ssumm+x1y2; endif
if (x1y3 < 0.0_quad_real) then; ssump=ssump-x1y3; else; ssumm=ssumm-x1y3; endif
if (x2y3 > 0.0_quad_real) then; ssump=ssump+x2y3; else; ssumm=ssumm+x2y3; endif
if (x2y1 < 0.0_quad_real) then; ssump=ssump-x2y1; else; ssumm=ssumm-x2y1; endif
if (x3y1 > 0.0_quad_real) then; ssump=ssump+x3y1; else; ssumm=ssumm+x3y1; endif
if (x3y2 < 0.0_quad_real) then; ssump=ssump-x3y2; else; ssumm=ssumm-x3y2; endif
det = ssump + ssumm

if (det == 0.0_refsoln_real) then
   errcode = 1
   return
else
   errcode = 0
endif

! if the request is to not divide by the determinant, only use its sign

if (local_no_det) then
   det = sign(1.0_quad_real,det)
endif

! compute the coordinates

! zeta(1) = (x*(y2-y3) + y*(x3-x2) + (x2*y3-x3*y2))/det

sump=0.0_quad_real; summ=0.0_quad_real
if (xy2 > 0.0_quad_real) then
   sump=sump+xy2
else
   summ=summ+xy2
endif
if (xy3 < 0.0_quad_real) then
   sump=sump-xy3
else
   summ=summ-xy3
endif
if (yx3 > 0.0_quad_real) then
   sump=sump+yx3
else
   summ=summ+yx3
endif
if (yx2 < 0.0_quad_real) then
   sump=sump-yx2
else
   summ=summ-yx2
endif
if (x2y3 > 0.0_quad_real) then; sump=sump+x2y3; else; summ=summ+x2y3; endif
if (x3y2 < 0.0_quad_real) then; sump=sump-x3y2; else; summ=summ-x3y2; endif
zeta(1) = sump/det + summ/det

! zeta(2) = (x*(y3-y1) + y*(x1-x3) + (x3*y1-x1*y3))/det

sump=0.0_quad_real; summ=0.0_quad_real
if (xy3 > 0.0_quad_real) then
   sump=sump+xy3
else
   summ=summ+xy3
endif
if (xy1 < 0.0_quad_real) then
   sump=sump-xy1
else
   summ=summ-xy1
endif
if (yx1 > 0.0_quad_real) then
   sump=sump+yx1
else
   summ=summ+yx1
endif
if (yx3 < 0.0_quad_real) then
   sump=sump-yx3
else
   summ=summ-yx3
endif
if (x3y1 > 0.0_quad_real) then; sump=sump+x3y1; else; summ=summ+x3y1; endif
if (x1y3 < 0.0_quad_real) then; sump=sump-x1y3; else; summ=summ-x1y3; endif
zeta(2) = sump/det + summ/det

! zeta(3) = (x*(y1-y2) + y*(x2-x1) + (x1*y2-x2*y1))/det

sump=0.0_quad_real; summ=0.0_quad_real
if (xy1 > 0.0_quad_real) then
   sump=sump+xy1
else
   summ=summ+xy1
endif
if (xy2 < 0.0_quad_real) then
   sump=sump-xy2
else
   summ=summ-xy2
endif
if (yx2 > 0.0_quad_real) then
   sump=sump+yx2
else
   summ=summ+yx2
endif
if (yx1 < 0.0_quad_real) then
   sump=sump-yx1
else
   summ=summ-yx1
endif
if (x1y2 > 0.0_quad_real) then; sump=sump+x1y2; else; summ=summ+x1y2; endif
if (x2y1 < 0.0_quad_real) then; sump=sump-x2y1; else; summ=summ-x2y1; endif
zeta(3) = sump/det + summ/det

! derivatives

if (present(zetax)) then
   zetax(1) = (y2-y3)/det
   zetax(2) = (y3-y1)/det
   zetax(3) = (y1-y2)/det
endif
if (present(zetay)) then
   zetay(1) = (x3-x2)/det
   zetay(2) = (x1-x3)/det
   zetay(3) = (x2-x1)/det
endif

end subroutine barycentric

!          ---------------
subroutine barycentric_tet(x,y,z,xv,yv,zv,zeta,errcode,zetax,zetay,zetaz)
!          ---------------

!----------------------------------------------------
! This function returns the barycentric coordinates of (x,y,z) in the
! tetrahedral element with vertices (xv,yv,zv).
!
! errcode: 0 success
!          1 degenerate element
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

real(refsoln_real), intent(in) :: x,y,z,xv(:),yv(:),zv(:)
real(refsoln_real), intent(out) :: zeta(:)
real(refsoln_real), intent(out), optional :: zetax(:), zetay(:), zetaz(:)
integer, intent(out) :: errcode
!----------------------------------------------------
! Local variables:

real(refsoln_real) :: A(3,3), b(3), det
integer :: ipiv(3), info, i
!----------------------------------------------------
! Begin executable code

A(1,1) = xv(1) - xv(4); A(1,2) = xv(2) - xv(4); A(1,3) = xv(3) - xv(4)
A(2,1) = yv(1) - yv(4); A(2,2) = yv(2) - yv(4); A(2,3) = yv(3) - yv(4)
A(3,1) = zv(1) - zv(4); A(3,2) = zv(2) - zv(4); A(3,3) = zv(3) - zv(4)

b(1) = x - xv(4)
b(2) = y - yv(4)
b(3) = z - zv(4)

! compute the barycentric coordinates of the point

det = A(1,1)*(A(2,2)*A(3,3)-A(3,2)*A(2,3)) - &
      A(2,1)*(A(1,2)*A(3,3)-A(3,2)*A(1,3)) + &
      A(3,1)*(A(1,2)*A(2,3)-A(2,2)*A(1,3))

if (det == 0.0_refsoln_real) then
   errcode = 1
   return
else
   errcode = 0
endif

! compute the coordinates

zeta(1) = (b(1)*(A(2,2)*A(3,3)-A(3,2)*A(2,3)) - &
           b(2)*(A(1,2)*A(3,3)-A(3,2)*A(1,3)) + &
           b(3)*(A(1,2)*A(2,3)-A(2,2)*A(1,3)))/det

zeta(2) = (A(1,1)*(b(2)*A(3,3)-b(3)*A(2,3)) - &
           A(2,1)*(b(1)*A(3,3)-b(3)*A(1,3)) + &
           A(3,1)*(b(1)*A(2,3)-b(2)*A(1,3)))/det

zeta(3) = (A(1,1)*(A(2,2)*b(3)-A(3,2)*b(2)) - &
           A(2,1)*(A(1,2)*b(3)-A(3,2)*b(1)) + &
           A(3,1)*(A(1,2)*b(2)-A(2,2)*b(1)))/det

zeta(4) = 1.0_refsoln_real - zeta(1) - zeta(2) - zeta(3)

! Derivatives

if (present(zetax)) then
   zetax(1) = (A(2,2)*A(3,3)-A(3,2)*A(2,3))/det
   zetax(2) = (A(3,1)*A(2,3)-A(2,1)*A(3,3))/det
   zetax(3) = (A(2,1)*A(3,2)-A(3,1)*A(2,2))/det
   zetax(4) = - zetax(1) - zetax(2) - zetax(3)
endif

if (present(zetay)) then
   zetay(1) = (A(3,2)*A(1,3)-A(1,2)*A(3,3))/det
   zetay(2) = (A(1,1)*A(3,3)-A(3,1)*A(1,3))/det
   zetay(3) = (A(3,1)*A(1,2)-A(1,1)*A(3,2))/det
   zetay(4) = - zetay(1) - zetay(2) - zetay(3)
endif

if (present(zetaz)) then
   zetaz(1) = (A(1,2)*A(2,3)-A(2,2)*A(1,3))/det
   zetaz(2) = (A(2,1)*A(1,3)-A(1,1)*A(2,3))/det
   zetaz(3) = (A(1,1)*A(2,2)-A(2,1)*A(1,2))/det
   zetaz(4) = - zetaz(1) - zetaz(2) - zetaz(3)
endif

end subroutine barycentric_tet

!          -------------------
subroutine evaluate_refsoln_1D(refsoln,x,elem,u,errcode,ux,soln_id)
!          -------------------

type(refsoln_type), intent(in) :: refsoln
real(refsoln_real), intent(in) :: x(:)
integer, intent(in) :: elem
real(refsoln_real), intent(out) :: u
real(refsoln_real), intent(out), optional :: ux
integer, intent(in), optional :: soln_id
integer, intent(out) :: errcode

! This function evaluates a 1D solution, and optionally first derivative,
! at the point x which is contained in element elem.
!
! errcode: 0 success
!          1 degenerate element

integer :: loc_soln_id
real(refsoln_real) :: basis(refsoln%degree+1), basisx(refsoln%degree+1)

! Determine which solution to use, if there is more than one

if (present(soln_id)) then
   loc_soln_id = soln_id
else
   loc_soln_id = 1
endif

! Evaluate the basis functions at the given point

if (present(ux)) then
   call basis1D(x(1), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:2,elem)), &
                refsoln%degree,basis,errcode,basisx)
else
   call basis1D(x(1), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:2,elem)), &
                refsoln%degree,basis,errcode)
endif
if (errcode /= 0) then
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: degenerate element"
   endif
   return
endif

! Evaluate the solution

u = sum(refsoln%mesh%solution(loc_soln_id, &
        refsoln%mesh%element_node(:,elem)) * basis)
if (present(ux)) then
   ux = sum(refsoln%mesh%solution(loc_soln_id, &
            refsoln%mesh%element_node(:,elem)) * basisx)
endif

end subroutine evaluate_refsoln_1D

!          -------------------
subroutine evaluate_refsoln_2D(refsoln,x,elem,u,errcode,ux,uy,soln_id)
!          -------------------

type(refsoln_type), intent(in) :: refsoln
real(refsoln_real), intent(in) :: x(:)
integer, intent(in) :: elem
real(refsoln_real), intent(out) :: u
real(refsoln_real), intent(out), optional :: ux
real(refsoln_real), intent(out), optional :: uy
integer, intent(in), optional :: soln_id
integer, intent(out) :: errcode

! This function evaluates a 2D solution, and optionally first derivatives,
! at the point x which is contained in element elem.
!
! errcode: 0 success
!          1 degenerate element

integer :: loc_soln_id, which
real(refsoln_real) :: basis(((refsoln%degree+1)*(refsoln%degree+2))/2), &
                      basisx(((refsoln%degree+1)*(refsoln%degree+2))/2), &
                      basisy(((refsoln%degree+1)*(refsoln%degree+2))/2)

errcode = 0
u = 0
if (present(ux)) ux = 0
if (present(uy)) uy = 0

! Determine which solution to use, if there is more than one

if (present(soln_id)) then
   loc_soln_id = soln_id
else
   loc_soln_id = 1
endif

! Determine which derivatives to include in basis evaluation

which = 0
if (present(ux)) which = 1
if (present(uy)) which = which + 2

! Evaluate the basis functions at the given point

select case(which)
case(0)
   call basis2D(x(1),x(2), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:3,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:3,elem)), &
                refsoln%degree,basis,errcode)
case(1)
   call basis2D(x(1),x(2), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:3,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:3,elem)), &
                refsoln%degree,basis,errcode,basisx)
case(2)
   call basis2D(x(1),x(2), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:3,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:3,elem)), &
                refsoln%degree,basis,errcode,basisy=basisy)
case(3)
   call basis2D(x(1),x(2), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:3,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:3,elem)), &
                refsoln%degree,basis,errcode,basisx,basisy)
end select
if (errcode /= 0) then
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: degenerate triangle"
   endif
   return
endif

! Evaluate the solution

u = sum(refsoln%mesh%solution(loc_soln_id, &
        refsoln%mesh%element_node(:,elem)) * basis)
if (present(ux)) then
   ux = sum(refsoln%mesh%solution(loc_soln_id, &
            refsoln%mesh%element_node(:,elem)) * basisx)
endif
if (present(uy)) then
   uy = sum(refsoln%mesh%solution(loc_soln_id, &
            refsoln%mesh%element_node(:,elem)) * basisy)
endif

if (refsoln%verbosity >= 5) then
   print *,"2D solution evaluation"
   print *," point",x
   print *," element",elem
   print *," xvert",refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:3,elem))
   print *," yvert",refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:3,elem))
   print *," basis",basis
   if (present(ux)) print *," basisx",basisx
   if (present(uy)) print *," basisy",basisy
   print *," node soln",refsoln%mesh%solution(loc_soln_id, &
                          refsoln%mesh%element_node(:,elem))
   print *," solution",u
   if (present(ux)) print *," x derivative",ux
   if (present(uy)) print *," y derivative",uy
endif

end subroutine evaluate_refsoln_2D

!          -------------------
subroutine evaluate_refsoln_3D(refsoln,x,elem,u,errcode,ux,uy,uz,soln_id)
!          -------------------

type(refsoln_type), intent(in) :: refsoln
real(refsoln_real), intent(in) :: x(:)
integer, intent(in) :: elem
real(refsoln_real), intent(out) :: u
real(refsoln_real), intent(out), optional :: ux
real(refsoln_real), intent(out), optional :: uy
real(refsoln_real), intent(out), optional :: uz
integer, intent(in), optional :: soln_id
integer, intent(out) :: errcode

! This function evaluates a 3D solution and optionally first derivatives at the
! point x which is contained in element elem.
!
! errcode: 0 success
!          1 degenerate element

integer :: loc_soln_id, which
real(refsoln_real) :: basis(((refsoln%degree+1)*(refsoln%degree+2)* &
                             (refsoln%degree+3))/6), &
                      basisx(((refsoln%degree+1)*(refsoln%degree+2)* &
                             (refsoln%degree+3))/6), &
                      basisy(((refsoln%degree+1)*(refsoln%degree+2)* &
                             (refsoln%degree+3))/6), &
                      basisz(((refsoln%degree+1)*(refsoln%degree+2)* &
                             (refsoln%degree+3))/6)

errcode = 0
u = 0
if (present(ux)) ux = 0
if (present(uy)) uy = 0
if (present(uz)) uz = 0

! Determine which solution to use, if there is more than one

if (present(soln_id)) then
   loc_soln_id = soln_id
else
   loc_soln_id = 1
endif

! Determine which derivatives to include in basis evaluation

which = 0
if (present(ux)) which = 1
if (present(uy)) which = which + 2
if (present(uz)) which = which + 4

! Evaluate the basis functions at the given point

select case(which)
case(0)
   call basis3D(x(1),x(2),x(3), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(3,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%degree,basis,errcode)
case(1)
   call basis3D(x(1),x(2),x(3), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(3,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%degree,basis,errcode,basisx)
case(2)
   call basis3D(x(1),x(2),x(3), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(3,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%degree,basis,errcode,basisy=basisy)
case(3)
   call basis3D(x(1),x(2),x(3), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(3,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%degree,basis,errcode,basisx,basisy)
case(4)
   call basis3D(x(1),x(2),x(3), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(3,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%degree,basis,errcode,basisz=basisz)
case(5)
   call basis3D(x(1),x(2),x(3), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(3,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%degree,basis,errcode,basisx,basisz=basisz)
case(6)
   call basis3D(x(1),x(2),x(3), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(3,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%degree,basis,errcode,basisy=basisy,basisz=basisz)
case(7)
   call basis3D(x(1),x(2),x(3), &
                refsoln%mesh%vertex(1,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(2,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%mesh%vertex(3,refsoln%mesh%element_node(1:4,elem)), &
                refsoln%degree,basis,errcode,basisx,basisy,basisz)
end select
if (errcode /= 0) then
   if (refsoln%verbosity > 0) then
      print *,"ERROR in evaluate_refsoln: degenerate tetrahedron"
   endif
   return
endif

! Evaluate the solution

u = sum(refsoln%mesh%solution(loc_soln_id, &
        refsoln%mesh%element_node(:,elem))*basis)
if (present(ux)) then
   ux = sum(refsoln%mesh%solution(loc_soln_id, &
            refsoln%mesh%element_node(:,elem))*basisx)
endif
if (present(uy)) then
   uy = sum(refsoln%mesh%solution(loc_soln_id, &
            refsoln%mesh%element_node(:,elem))*basisy)
endif
if (present(uz)) then
   uz = sum(refsoln%mesh%solution(loc_soln_id, &
            refsoln%mesh%element_node(:,elem))*basisz)
endif

end subroutine evaluate_refsoln_3D

!          -------
subroutine basis1D(x,xvert,degree,basis,errcode,basisx)
!          -------

real(refsoln_real), intent(in) :: x
real(refsoln_real), intent(in) :: xvert(2)
integer, intent(in) :: degree
! Dimension degree+1
real(refsoln_real), intent(out) :: basis(:)
real(refsoln_real), intent(out), optional :: basisx(:)
integer, intent(out) :: errcode

! errcode: 0 success
!          1 degenerate element

real(refsoln_real) :: xnode(degree+1), rdegree, alpha
integer :: i, j

if (xvert(1) == xvert(2)) then
   errcode = 1
   return
else
   errcode = 0
endif

xnode(1) = xvert(1)
xnode(2) = xvert(2)
rdegree = degree
do i=1,degree-1
   alpha = i/rdegree
   xnode(2+i) = (1-alpha)*xvert(1) + alpha*xvert(2)
end do

do i=1,degree+1
   basis(i) = 1
   if (present(basisx)) basisx(i) = 1
   do j=1,degree+1
      if (i==j) cycle
      if (present(basisx)) then
         basisx(i) = basis(i)/(xnode(i)-xnode(j)) + &
                     basisx(i)*(x-xnode(j))/(xnode(i)-xnode(j))
      endif
      basis(i) = basis(i)*(x-xnode(j))/(xnode(i)-xnode(j))
   end do
end do

end subroutine basis1D

!          -------
subroutine basis2D(x,y,xvert,yvert,degree,basis,errcode,basisx,basisy)
!          -------

real(refsoln_real), intent(in) :: x, y
real(refsoln_real), intent(in) :: xvert(3), yvert(3)
integer, intent(in) :: degree
! Dimension ((degree+1)*(degree+2))/2
real(refsoln_real), intent(out) :: basis(:)
real(refsoln_real), intent(out), optional :: basisx(:), basisy(:)
integer, intent(out) :: errcode

real(refsoln_real) :: zeta(3), zetax(3), zetay(3), rdegree, alpha
integer :: i, j, k, b, v1, v2

! Evaluate the Lagrange basis functions using the barycentric coordinates
! as described in www.cs.rpi.edu/~flaherje/pdf/fea4.pdf
!
! errcode: 0 success
!          1 degenerate element

! The order of the nodes, using degree 4 as an example, is
!         1
!       4   7
!     5  13   8
!   6  14  15   9
! 2  10  11  12  3

basis = 0
if (present(basisx)) basisx = 0
if (present(basisy)) basisy = 0

call barycentric(x,y,xvert,yvert,zeta,errcode,zetax,zetay)
if (errcode /= 0) return

rdegree = degree

! Vertex basis functions

do b=1,3
   basis(b) = zeta(b)
   if (present(basisx)) basisx(b) = zetax(b)
   if (present(basisy)) basisy(b) = zetay(b)
   alpha = 1
   do j=1,degree-1
      if (present(basisx)) then
         basisx(b) = basis(b)*zetax(b) + &
                     basisx(b)*(zeta(b)-j/rdegree)
      endif
      if (present(basisy)) then
         basisy(b) = basis(b)*zetay(b) + &
                     basisy(b)*(zeta(b)-j/rdegree)
      endif
      basis(b) = basis(b)*(zeta(b)-j/rdegree)
      alpha = alpha*(1-j/rdegree)
   end do
   basis(b) = basis(b)/alpha
   if (present(basisx)) basisx(b) = basisx(b)/alpha
   if (present(basisy)) basisy(b) = basisy(b)/alpha
end do

! Edge basis functions

b=3
do i=1,3 ! each i is an edge of the triangle
   select case(i)
   case (1)
      v1 = 1
      v2 = 2
   case (2)
      v1 = 1
      v2 = 3
   case (3)
      v1 = 2
      v2 = 3
   end select
   do j=1,degree-1
      b = b+1
      basis(b) = zeta(v1)*zeta(v2)
      if (present(basisx)) then
         basisx(b) = zeta(v1)*zetax(v2) + zetax(v1)*zeta(v2)
      endif
      if (present(basisy)) then
         basisy(b) = zeta(v1)*zetay(v2) + zetay(v1)*zeta(v2)
      endif
      alpha = (rdegree-j)/rdegree
      alpha = alpha*j/rdegree
      do k=1,degree-1-j
         if (present(basisx)) then
            basisx(b) = basis(b)*zetax(v1) + &
                        basisx(b)*(zeta(v1)-k/rdegree)
         endif
         if (present(basisy)) then
            basisy(b) = basis(b)*zetay(v1) + &
                        basisy(b)*(zeta(v1)-k/rdegree)
         endif
         basis(b) = basis(b)*(zeta(v1)-k/rdegree)
         alpha = alpha*(rdegree-j-k)/rdegree
      end do
      do k=1,j-1
         if (present(basisx)) then
            basisx(b) = basis(b)*zetax(v2) + &
                        basisx(b)*(zeta(v2)-k/rdegree)
         endif
         if (present(basisy)) then
            basisy(b) = basis(b)*zetay(v2) + &
                        basisy(b)*(zeta(v2)-k/rdegree)
         endif
         basis(b) = basis(b)*(zeta(v2)-k/rdegree)
         alpha = alpha*(j-k)/rdegree
      end do
      basis(b) = basis(b)/alpha
      if (present(basisx)) basisx(b) = basisx(b)/alpha
      if (present(basisy)) basisy(b) = basisy(b)/alpha
   end do
end do

! Interior basis functions

do i=1,degree-2 ! each i is a line of nodes parallel to edge 2-3
   do j=1,i
      b = b+1
      basis(b) = zeta(1)*zeta(2)*zeta(3)
      if (present(basisx)) then
         basisx(b) = zeta(1)*zeta(2)*zetax(3) + &
                     zeta(1)*zetax(2)*zeta(3) + &
                     zetax(1)*zeta(2)*zeta(3)
      endif
      if (present(basisy)) then
         basisy(b) = zeta(1)*zeta(2)*zetay(3) + &
                     zeta(1)*zetay(2)*zeta(3) + &
                     zetay(1)*zeta(2)*zeta(3)
      endif
      alpha = (rdegree-i-1)/rdegree
      alpha = alpha*(i-j+1)/rdegree
      alpha = alpha*j/rdegree
      do k=1,degree-2-i
         if (present(basisx)) then
            basisx(b) = basis(b)*zetax(1) + &
                        basisx(b)*(zeta(1)-k/rdegree)
         endif
         if (present(basisy)) then
            basisy(b) = basis(b)*zetay(1) + &
                        basisy(b)*(zeta(1)-k/rdegree)
         endif
         basis(b) = basis(b)*(zeta(1)-k/rdegree)
         alpha = alpha*(rdegree-i-1-k)/rdegree
      end do
      do k=1,i-j
         if (present(basisx)) then
            basisx(b) = basis(b)*zetax(2) + &
                        basisx(b)*(zeta(2)-k/rdegree)
         endif
         if (present(basisy)) then
            basisy(b) = basis(b)*zetay(2) + &
                        basisy(b)*(zeta(2)-k/rdegree)
         endif
         basis(b) = basis(b)*(zeta(2)-k/rdegree)
         alpha = alpha*(i-j+1-k)/rdegree
      end do
      do k=1,j-1
         if (present(basisx)) then
            basisx(b) = basis(b)*zetax(3) + &
                        basisx(b)*(zeta(3)-k/rdegree)
         endif
         if (present(basisy)) then
            basisy(b) = basis(b)*zetay(3) + &
                        basisy(b)*(zeta(3)-k/rdegree)
         endif
         basis(b) = basis(b)*(zeta(3)-k/rdegree)
         alpha = alpha*(j-k)/rdegree
      end do
      basis(b) = basis(b)/alpha
      if (present(basisx)) basisx(b) = basisx(b)/alpha
      if (present(basisy)) basisy(b) = basisy(b)/alpha
   end do
end do

end subroutine basis2D

!          -------
subroutine basis3D(x,y,z,xvert,yvert,zvert,degree,basis,errcode,basisx,basisy, &
                   basisz)
!          -------

real(refsoln_real), intent(in) :: x, y, z
real(refsoln_real), intent(in) :: xvert(4),yvert(4),zvert(4)
integer, intent(in) :: degree
! Dimension ((degree+1)*(degree+2)*(degree+3))/6
real(refsoln_real), intent(out) :: basis(:)
real(refsoln_real), intent(out), optional :: basisx(:), basisy(:), basisz(:)
integer, intent(out) :: errcode

real(refsoln_real) :: zeta(4), zetax(4), zetay(4), zetaz(4), rdegree, alpha
integer :: i, j, k, l, b, v1, v2, v3

! Evaluate the Lagrange basis functions using the barycentric coordinates.
!
! errcode: 0 success
!          1 degenerate element

basis = 0
if (present(basisx)) basisx = 0
if (present(basisy)) basisy = 0
if (present(basisz)) basisz = 0

call barycentric_tet(x,y,z,xvert,yvert,zvert,zeta,errcode,zetax,zetay,zetaz)
if (errcode /= 0) return

rdegree = degree

! Vertex basis functions

do b=1,4
   basis(b) = zeta(b)
   if (present(basisx)) basisx(b) = zetax(b)
   if (present(basisy)) basisy(b) = zetay(b)
   if (present(basisz)) basisz(b) = zetaz(b)
   alpha = 1
   do j=1,degree-1
      if (present(basisx)) then
         basisx(b) = basis(b)*zetax(b) + &
                     basisx(b)*(zeta(b)-j/rdegree)
      endif
      if (present(basisy)) then
         basisy(b) = basis(b)*zetay(b) + &
                     basisy(b)*(zeta(b)-j/rdegree)
      endif
      if (present(basisz)) then
         basisz(b) = basis(b)*zetaz(b) + &
                     basisz(b)*(zeta(b)-j/rdegree)
      endif
      basis(b) = basis(b)*(zeta(b)-j/rdegree)
      alpha = alpha*(1-j/rdegree)
   end do
   basis(b) = basis(b)/alpha
   if (present(basisx)) basisx(b) = basisx(b)/alpha
   if (present(basisy)) basisy(b) = basisy(b)/alpha
   if (present(basisz)) basisz(b) = basisz(b)/alpha
end do

! Edge basis functions

b=4
do i=1,6 ! each i is an edge of the tetrahedron
   select case(i)
   case (1)
      v1 = 1
      v2 = 2
   case (2)
      v1 = 1
      v2 = 3
   case (3)
      v1 = 1
      v2 = 4
   case (4)
      v1 = 2
      v2 = 3
   case (5)
      v1 = 2
      v2 = 4
   case (6)
      v1 = 3
      v2 = 4
   end select
   do j=1,degree-1
      b = b+1
      basis(b) = zeta(v1)*zeta(v2)
      if (present(basisx)) then
         basisx(b) = zeta(v1)*zetax(v2) + zetax(v1)*zeta(v2)
      endif
      if (present(basisy)) then
         basisy(b) = zeta(v1)*zetay(v2) + zetay(v1)*zeta(v2)
      endif
      if (present(basisz)) then
         basisz(b) = zeta(v1)*zetaz(v2) + zetaz(v1)*zeta(v2)
      endif
      alpha = (rdegree-j)/rdegree
      alpha = alpha*j/rdegree
      do k=1,degree-1-j
         if (present(basisx)) then
            basisx(b) = basis(b)*zetax(v1) + &
                        basisx(b)*(zeta(v1)-k/rdegree)
         endif
         if (present(basisy)) then
            basisy(b) = basis(b)*zetay(v1) + &
                        basisy(b)*(zeta(v1)-k/rdegree)
         endif
         if (present(basisz)) then
            basisz(b) = basis(b)*zetaz(v1) + &
                        basisz(b)*(zeta(v1)-k/rdegree)
         endif
         basis(b) = basis(b)*(zeta(v1)-k/rdegree)
         alpha = alpha*(rdegree-j-k)/rdegree
      end do
      do k=1,j-1
         if (present(basisx)) then
            basisx(b) = basis(b)*zetax(v2) + &
                        basisx(b)*(zeta(v2)-k/rdegree)
         endif
         if (present(basisy)) then
            basisy(b) = basis(b)*zetay(v2) + &
                        basisy(b)*(zeta(v2)-k/rdegree)
         endif
         if (present(basisz)) then
            basisz(b) = basis(b)*zetaz(v2) + &
                        basisz(b)*(zeta(v2)-k/rdegree)
         endif
         basis(b) = basis(b)*(zeta(v2)-k/rdegree)
         alpha = alpha*(j-k)/rdegree
      end do
      basis(b) = basis(b)/alpha
      if (present(basisx)) basisx(b) = basisx(b)/alpha
      if (present(basisy)) basisy(b) = basisy(b)/alpha
      if (present(basisz)) basisz(b) = basisz(b)/alpha
   end do
end do

! Face basis functions

do l=1,4 ! each l is a face of the tetrahedron
   select case(l)
   case (1)
      v1 = 1
      v2 = 2
      v3 = 3
   case (2)
      v1 = 1
      v2 = 2
      v3 = 4
   case (3)
      v1 = 1
      v2 = 3
      v3 = 4
   case (4)
      v1 = 2
      v2 = 3
      v3 = 4
   end select
   do i=1,degree-2 ! each i is a line of nodes parallel to edge v2-v3
      do j=1,i
         b = b+1
         basis(b) = zeta(v1)*zeta(v2)*zeta(v3)
         if (present(basisx)) then
            basisx(b) = zeta(v1)*zeta(v2)*zetax(v3) + &
                        zeta(v1)*zetax(v2)*zeta(v3) + &
                        zetax(v1)*zeta(v2)*zeta(v3)
         endif
         if (present(basisy)) then
            basisy(b) = zeta(v1)*zeta(v2)*zetay(v3) + &
                        zeta(v1)*zetay(v2)*zeta(v3) + &
                        zetay(v1)*zeta(v2)*zeta(v3)
         endif
         if (present(basisz)) then
            basisz(b) = zeta(v1)*zeta(v2)*zetaz(v3) + &
                        zeta(v1)*zetaz(v2)*zeta(v3) + &
                        zetaz(v1)*zeta(v2)*zeta(v3)
         endif
         alpha = (rdegree-i-1)/rdegree
         alpha = alpha*(i-j+1)/rdegree
         alpha = alpha*j/rdegree
         do k=1,degree-2-i
            if (present(basisx)) then
               basisx(b) = basis(b)*zetax(v1) + &
                           basisx(b)*(zeta(v1)-k/rdegree)
            endif
            if (present(basisy)) then
               basisy(b) = basis(b)*zetay(v1) + &
                           basisy(b)*(zeta(v1)-k/rdegree)
            endif
            if (present(basisz)) then
               basisz(b) = basis(b)*zetaz(v1) + &
                           basisz(b)*(zeta(v1)-k/rdegree)
            endif
            basis(b) = basis(b)*(zeta(v1)-k/rdegree)
            alpha = alpha*(rdegree-i-1-k)/rdegree
         end do
         do k=1,i-j
            if (present(basisx)) then
               basisx(b) = basis(b)*zetax(v2) + &
                           basisx(b)*(zeta(v2)-k/rdegree)
            endif
            if (present(basisy)) then
               basisy(b) = basis(b)*zetay(v2) + &
                           basisy(b)*(zeta(v2)-k/rdegree)
            endif
            if (present(basisz)) then
               basisz(b) = basis(b)*zetaz(v2) + &
                           basisz(b)*(zeta(v2)-k/rdegree)
            endif
            basis(b) = basis(b)*(zeta(v2)-k/rdegree)
            alpha = alpha*(i-j+1-k)/rdegree
         end do
         do k=1,j-1
            if (present(basisx)) then
               basisx(b) = basis(b)*zetax(v3) + &
                           basisx(b)*(zeta(v3)-k/rdegree)
            endif
            if (present(basisy)) then
               basisy(b) = basis(b)*zetay(v3) + &
                           basisy(b)*(zeta(v3)-k/rdegree)
            endif
            if (present(basisz)) then
               basisz(b) = basis(b)*zetaz(v3) + &
                           basisz(b)*(zeta(v3)-k/rdegree)
            endif
            basis(b) = basis(b)*(zeta(v3)-k/rdegree)
            alpha = alpha*(j-k)/rdegree
         end do
         basis(b) = basis(b)/alpha
         if (present(basisx)) basisx(b) = basisx(b)/alpha
         if (present(basisy)) basisy(b) = basisy(b)/alpha
         if (present(basisz)) basisz(b) = basisz(b)/alpha
      end do
   end do
end do

! Interior basis functions

do l=1,degree-3 ! each l is a plane of nodes parallel to face 2-3-4
   do i=1,l ! each i is a line of nodes parallel to edge 3-4
      do j=1,i
         b = b+1
         basis(b) = zeta(1)*zeta(2)*zeta(3)*zeta(4)
         if (present(basisx)) then
            basisx(b) = zeta(1)*zeta(2)*zeta(3)*zetax(4) + &
                        zeta(1)*zeta(2)*zetax(3)*zeta(4) + &
                        zeta(1)*zetax(2)*zeta(3)*zeta(4) + &
                        zetax(1)*zeta(2)*zeta(3)*zeta(4)
         endif
         if (present(basisy)) then
            basisy(b) = zeta(1)*zeta(2)*zeta(3)*zetay(4) + &
                        zeta(1)*zeta(2)*zetay(3)*zeta(4) + &
                        zeta(1)*zetay(2)*zeta(3)*zeta(4) + &
                        zetay(1)*zeta(2)*zeta(3)*zeta(4)
         endif
         if (present(basisz)) then
            basisz(b) = zeta(1)*zeta(2)*zeta(3)*zetaz(4) + &
                        zeta(1)*zeta(2)*zetaz(3)*zeta(4) + &
                        zeta(1)*zetaz(2)*zeta(3)*zeta(4) + &
                        zetaz(1)*zeta(2)*zeta(3)*zeta(4)
         endif
         alpha = (rdegree-2-l)/rdegree
         alpha = alpha*(l-i+1)/rdegree
         alpha = alpha*(i-j+1)/rdegree
         alpha = alpha*j/rdegree
         do k=1,degree-3-l
            if (present(basisx)) then
               basisx(b) = basis(b)*zetax(1) + &
                           basisx(b)*(zeta(1)-k/rdegree)
            endif
            if (present(basisy)) then
               basisy(b) = basis(b)*zetay(1) + &
                           basisy(b)*(zeta(1)-k/rdegree)
            endif
            if (present(basisz)) then
               basisz(b) = basis(b)*zetaz(1) + &
                           basisz(b)*(zeta(1)-k/rdegree)
            endif
            basis(b) = basis(b)*(zeta(1)-k/rdegree)
            alpha = alpha*(rdegree-2-l-k)/rdegree
         end do
         do k=1,l-i
            if (present(basisx)) then
               basisx(b) = basis(b)*zetax(2) + &
                           basisx(b)*(zeta(2)-k/rdegree)
            endif
            if (present(basisy)) then
               basisy(b) = basis(b)*zetay(2) + &
                           basisy(b)*(zeta(2)-k/rdegree)
            endif
            if (present(basisz)) then
               basisz(b) = basis(b)*zetaz(2) + &
                           basisz(b)*(zeta(2)-k/rdegree)
            endif
            basis(b) = basis(b)*(zeta(2)-k/rdegree)
            alpha = alpha*(l-i+1-k)/rdegree
         end do
         do k=1,i-j
            if (present(basisx)) then
               basisx(b) = basis(b)*zetax(3) + &
                           basisx(b)*(zeta(3)-k/rdegree)
            endif
            if (present(basisy)) then
               basisy(b) = basis(b)*zetay(3) + &
                           basisy(b)*(zeta(3)-k/rdegree)
            endif
            if (present(basisz)) then
               basisz(b) = basis(b)*zetaz(3) + &
                           basisz(b)*(zeta(3)-k/rdegree)
            endif
            basis(b) = basis(b)*(zeta(3)-k/rdegree)
            alpha = alpha*(i-j+1-k)/rdegree
         end do
         do k=1,j-1
            if (present(basisx)) then
               basisx(b) = basis(b)*zetax(4) + &
                           basisx(b)*(zeta(4)-k/rdegree)
            endif
            if (present(basisy)) then
               basisy(b) = basis(b)*zetay(4) + &
                           basisy(b)*(zeta(4)-k/rdegree)
            endif
            if (present(basisz)) then
               basisz(b) = basis(b)*zetaz(4) + &
                           basisz(b)*(zeta(4)-k/rdegree)
            endif
            basis(b) = basis(b)*(zeta(4)-k/rdegree)
            alpha = alpha*(j-k)/rdegree
         end do
         basis(b) = basis(b)/alpha
         if (present(basisx)) basisx(b) = basisx(b)/alpha
         if (present(basisy)) basisy(b) = basisy(b)/alpha
         if (present(basisz)) basisz(b) = basisz(b)/alpha
      end do
   end do
end do

end subroutine basis3D

end module AMRrefsoln
