program example

! Example of using the AMRrefsoln module.  For a full list of the functions
! available in AMRrefsoln and their interfaces, see the comments at the
! beginning of AMRrefsoln.f90.

use AMRrefsoln
implicit none

type(refsoln_type) :: ref ! The reference solution object

integer, parameter :: NPOINT = 4
real(refsoln_real) :: x(2,NPOINT**2), u(NPOINT**2), xmin(2), xmax(2)
integer :: err, dim, nvert, nelem
integer :: i, j, isub

! Create the reference solution object

call create_refsoln(ref)

! Read a .nrs (NIST reference solution) file into the reference solution object

call read_refsoln(ref,"refsoln2D.nrs",err)
call checkerr("read_refsoln") ! checkerr is defined at the end of this file

! Verify that the .nrs file has a 2D reference solution

dim = get_refsoln_dimension(ref,err)
call checkerr("get_refsoln_dimension")
if (dim /= 2) then
   print *,"The nrs file does not have a 2D problem"
   call abort ! abort is defined at the end of this file
endif

! Print the number of vertices and elements in the mesh

nvert = get_refsoln_number_of_vertices(ref,err)
call checkerr("get_refsoln_number_of_vertices")
nelem = get_refsoln_number_of_elements(ref,err)
call checkerr("get_refsoln_number_of_elements")
print *,"The reference solution grid has ",nvert," vertices"
print *,"                            and ",nelem," elements"

! Create a NPOINT X NPOINT uniform grid of points in the bounding box that
! encloses the domain of the reference solution

call get_refsoln_bounding_box(ref,xmin,xmax,err)
call checkerr("get_refsoln_bounding_box")
isub = 0
do i=1,NPOINT
   do j=1,NPOINT
      isub = isub + 1
      x(1,isub) = xmin(1) + (xmax(1)-xmin(1))*(i-1)/(NPOINT-1.0_refsoln_real)
      x(2,isub) = xmin(2) + (xmax(2)-xmin(2))*(j-1)/(NPOINT-1.0_refsoln_real)
   end do
end do

! Evaluate the reference solution at the points just created

call evaluate_refsoln(ref,x,u,err)
call checkerr("evaluate_refsoln")

! Print the evaluated solution

print *
print *,"Reference solution at selected points; x y refsoln"
print *
do i=1,NPOINT**2
   print *,x(1,i), x(2,i), u(i)
end do

! Destroy the reference solution object

call destroy_refsoln(ref,err)
call checkerr("destroy_refsoln")

contains

subroutine checkerr(routine)
character(len=*), intent(in) :: routine
if (err /= 0) then
   print *,trim(routine)," returned error code",err
   call abort
endif
end subroutine checkerr

subroutine abort
call destroy_refsoln(ref,err)
stop
end subroutine abort

end program example
